# SPDX-License-Identifier: Apache-2.0

from ..common._apply_operation import (
    apply_concat, apply_abs,
    apply_add, apply_mul, apply_div)
try:
    from ..common._apply_operation import apply_less
except ImportError:
    # onnxconverter-common is too old
    apply_less = None
from ..common.data_types import guess_proto_type
from ..common._registration import register_converter
from ..proto import onnx_proto
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_ovr_decision_function(
        scope: Scope, operator: Operator,
        container: ModelComponentContainer):

    # Applies _ovr_decision_function.
    # See https://github.com/scikit-learn/scikit-learn/blob/
    # master/sklearn/utils/multiclass.py#L407:
    # ::
    #     def _ovr_decision_function(predictions, confidences, n_classes):
    #
    #         n_samples = predictions.shape[0]
    #         votes = np.zeros((n_samples, n_classes))
    #         sum_of_confidences = np.zeros((n_samples, n_classes))
    #         k = 0
    #         for i in range(n_classes):
    #             for j in range(i + 1, n_classes):
    #                 sum_of_confidences[:, i] -= confidences[:, k]
    #                 sum_of_confidences[:, j] += confidences[:, k]
    #                 votes[predictions[:, k] == 0, i] += 1
    #                 votes[predictions[:, k] == 1, j] += 1
    #                 k += 1
    #         transformed_confidences = (
    #             sum_of_confidences / (3 * (np.abs(sum_of_confidences) + 1)))
    #         return votes + transformed_confidences
    proto_dtype = guess_proto_type(operator.inputs[1].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator

    cst3 = scope.get_unique_variable_name('cst3')
    container.add_initializer(cst3, proto_dtype, [], [3])
    cst1 = scope.get_unique_variable_name('cst1')
    container.add_initializer(cst1, proto_dtype, [], [1])

    iprediction = operator.inputs[0].full_name
    score_name = operator.inputs[1].full_name

    n_classes = len(op.classes_)
    sumc_name = [scope.get_unique_variable_name('svcsumc_%d' % i)
                 for i in range(n_classes)]
    vote_name = [scope.get_unique_variable_name('svcvote_%d' % i)
                 for i in range(n_classes)]
    sumc_add = {n: [] for n in sumc_name}
    vote_add = {n: [] for n in vote_name}
    k = 0
    for i in range(n_classes):
        for j in range(i + 1, n_classes):
            ind = scope.get_unique_variable_name('Cind_%d' % k)
            container.add_initializer(
                ind, onnx_proto.TensorProto.INT64, [], [k])

            # confidences
            ext = scope.get_unique_variable_name('Csvc_%d' % k)
            container.add_node(
                'ArrayFeatureExtractor', [score_name, ind],
                ext, op_domain='ai.onnx.ml')
            sumc_add[sumc_name[j]].append(ext)

            neg = scope.get_unique_variable_name('Cneg_%d' % k)
            container.add_node('Neg', ext, neg, op_domain='', op_version=6)
            sumc_add[sumc_name[i]].append(neg)

            # votes
            ext = scope.get_unique_variable_name('Vsvcv_%d' % k)
            container.add_node(
                'ArrayFeatureExtractor', [iprediction, ind],
                ext, op_domain='ai.onnx.ml')
            vote_add[vote_name[j]].append(ext)

            neg = scope.get_unique_variable_name('Vnegv_%d' % k)
            container.add_node('Neg', ext, neg, op_domain='', op_version=6)
            neg1 = scope.get_unique_variable_name('Vnegv1_%d' % k)
            apply_add(scope, [neg, cst1], neg1, container, broadcast=1,
                      operator_name='AddCl_%d_%d' % (i, j))
            vote_add[vote_name[i]].append(neg1)

            # next
            k += 1

    for k, v in sumc_add.items():
        name = scope.get_unique_operator_name('Sum')
        container.add_node(
            'Sum', v, k, op_domain='', name=name, op_version=8)
    for k, v in vote_add.items():
        name = scope.get_unique_operator_name('Sum')
        container.add_node(
            'Sum', v, k, op_domain='', name=name, op_version=8)

    conc = scope.get_unique_variable_name('Csvcconc')
    apply_concat(scope, sumc_name, conc, container, axis=1)
    conc_vote = scope.get_unique_variable_name('Vsvcconcv')
    apply_concat(scope, vote_name, conc_vote, container, axis=1)

    conc_abs = scope.get_unique_variable_name('Cabs')
    apply_abs(scope, conc, conc_abs, container)

    conc_abs1 = scope.get_unique_variable_name('Cconc_abs1')
    apply_add(scope, [conc_abs, cst1], conc_abs1, container, broadcast=1,
              operator_name='AddF0')
    conc_abs3 = scope.get_unique_variable_name('Cconc_abs3')
    apply_mul(scope, [conc_abs1, cst3], conc_abs3, container, broadcast=1)

    final = scope.get_unique_variable_name('Csvcfinal')
    apply_div(
        scope, [conc, conc_abs3], final, container, broadcast=0)

    output_name = operator.outputs[0].full_name
    apply_add(
        scope, [conc_vote, final], output_name, container, broadcast=0,
        operator_name='AddF1')


register_converter('SklearnOVRDecisionFunction',
                   convert_sklearn_ovr_decision_function)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common.data_types import (
    Int64TensorType, StringTensorType, Int32TensorType,
    FloatTensorType, DoubleTensorType
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_one_hot_encoder(scope: Scope, operator: Operator,
                                    container: ModelComponentContainer):
    """
    Converts *OneHotEncoder* into ONNX.
    It supports multiple inputs of types
    string or int64.
    """
    ohe_op = operator.raw_operator

    if len(operator.inputs) > 1:
        all_shapes = [inp.type.shape[1] for inp in operator.inputs]
        if any(map(lambda x: not isinstance(x, int) or x < 1, all_shapes)):
            raise RuntimeError(
                "Shapes must be known when OneHotEncoder is converted. "
                "There are {} inputs with the following number of columns "
                "{}.".format(len(operator.inputs), all_shapes))
        total = sum(all_shapes)
        if total != len(ohe_op.categories_):
            raise RuntimeError(
                "Mismatch between the number of sets of categories {} and "
                "the total number of inputs columns {}.".format(
                    len(ohe_op.categories_), total))

        enum_cats = []
        index_inputs = 0

        for index, cats in enumerate(ohe_op.categories_):
            while sum(all_shapes[:index_inputs+1]) <= index:
                index_inputs += 1
            index_in_input = index - sum(all_shapes[:index_inputs])

            inp = operator.inputs[index_inputs]
            if not isinstance(
                    inp.type,
                    (Int64TensorType, StringTensorType, Int32TensorType,
                     FloatTensorType, DoubleTensorType)):
                raise NotImplementedError(
                    "{} input datatype not yet supported. "
                    "You may raise an issue at "
                    "https://github.com/onnx/sklearn-onnx/issues"
                    "".format(type(inp.type)))

            if all_shapes[index_inputs] == 1:
                assert index_in_input == 0
                afeat = False
            else:
                afeat = True
            enum_cats.append(
                (afeat, index_in_input, inp.full_name, cats, inp.type))
    else:
        inp = operator.inputs[0]
        enum_cats = [(True, i, inp.full_name, cats, inp.type)
                     for i, cats in enumerate(ohe_op.categories_)]

    result, categories_len = [], 0
    for index, enum_c in enumerate(enum_cats):
        afeat, index_in, name, categories, inp_type = enum_c
        container.debug(
            "[conv.OneHotEncoder] cat %r/%r name=%r type=%r",
            index + 1, len(enum_cats), name, inp_type)
        if len(categories) == 0:
            continue
        if afeat:
            index_name = scope.get_unique_variable_name(
                name + str(index_in))
            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64, [], [index_in])
            out_name = scope.get_unique_variable_name(
                name + str(index_in))
            container.add_node(
                'ArrayFeatureExtractor', [name, index_name],
                out_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
            name = out_name

        attrs = {'name': scope.get_unique_operator_name('OneHotEncoder')}
        attrs['zeros'] = 1 if ohe_op.handle_unknown == 'ignore' else 0

        if isinstance(inp_type, (Int64TensorType, Int32TensorType)):
            attrs['cats_int64s'] = categories.astype(np.int64)
        elif isinstance(inp_type, StringTensorType):
            attrs['cats_strings'] = np.array(
                [str(s).encode('utf-8') for s in categories])
        elif isinstance(inp_type, (FloatTensorType, DoubleTensorType)):
            # The converter checks that categories can be casted into
            # integers. String is not allowed here.
            # Input type is casted into int64.
            for c in categories:
                try:
                    ci = int(c)
                except TypeError:
                    raise RuntimeError(
                        "Category '{}' cannot be casted into int.".format(c))
                if ci != c:
                    raise RuntimeError(
                        "Category %r is not an int64. "
                        "The converter only supports string and int64 "
                        "categories not %r." % (c, type(c)))
            attrs['cats_int64s'] = categories.astype(np.int64)
        else:
            raise RuntimeError(
                "Input type {} is not supported for OneHotEncoder. "
                "Ideally, it should either be integer or strings.".format(
                    inp_type))

        ohe_output = scope.get_unique_variable_name(name + 'out')

        if 'cats_int64s' in attrs:
            # Let's cast this input in int64.
            cast_feature = scope.get_unique_variable_name(name + 'cast')
            apply_cast(scope, name, cast_feature, container,
                       to=onnx_proto.TensorProto.INT64)
            name = cast_feature

        container.add_node('OneHotEncoder', name,
                           ohe_output, op_domain='ai.onnx.ml',
                           **attrs)
        if hasattr(ohe_op, 'drop_idx_') and ohe_op.drop_idx_ is not None:
            extracted_outputs_name = scope.get_unique_variable_name(
                'extracted_outputs')
            indices_to_keep_name = scope.get_unique_variable_name(
                'indices_to_keep')
            indices_to_keep = np.delete(
                np.arange(len(categories)), ohe_op.drop_idx_[index])
            container.add_initializer(
                indices_to_keep_name, onnx_proto.TensorProto.INT64,
                indices_to_keep.shape, indices_to_keep)
            container.add_node(
                'ArrayFeatureExtractor', [ohe_output, indices_to_keep_name],
                extracted_outputs_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
            ohe_output, categories = extracted_outputs_name, indices_to_keep

        result.append(ohe_output)
        categories_len += len(categories)

    concat_result_name = scope.get_unique_variable_name('concat_result')
    apply_concat(scope, result, concat_result_name, container, axis=2)

    reshape_input = concat_result_name
    if np.issubdtype(ohe_op.dtype, np.signedinteger):
        reshape_input = scope.get_unique_variable_name('cast')
        apply_cast(scope, concat_result_name, reshape_input,
                   container, to=onnx_proto.TensorProto.INT64)
    apply_reshape(scope, reshape_input, operator.output_full_names,
                  container, desired_shape=(-1, categories_len))


register_converter('SklearnOneHotEncoder', convert_sklearn_one_hot_encoder)



# SPDX-License-Identifier: Apache-2.0


# Rather than using ONNX protobuf definition throughout our codebase,
# we import ONNX protobuf definition here so that we can conduct quick
# fixes by overwriting ONNX functions without changing any lines
# elsewhere.
from onnx import onnx_pb as onnx_proto  # noqa
from onnx import defs  # noqa

# Overwrite the make_tensor defined in onnx.helper because of a bug
# (string tensor get assigned twice)
from onnx import mapping
from onnx.onnx_pb import TensorProto, ValueInfoProto  # noqa
try:
    from onnx.onnx_pb import SparseTensorProto  # noqa
except ImportError:
    # onnx is too old.
    pass
from onnx.helper import split_complex_to_pairs


def make_tensor_fixed(name, data_type, dims, vals, raw=False):
    '''
    Make a TensorProto with specified arguments.  If raw is False, this
    function will choose the corresponding proto field to store the
    values based on data_type. If raw is True, use "raw_data" proto
    field to store the values, and values should be of type bytes in
    this case.
    '''
    tensor = TensorProto()
    tensor.data_type = data_type
    tensor.name = name

    if (data_type == TensorProto.COMPLEX64 or
            data_type == TensorProto.COMPLEX128):
        vals = split_complex_to_pairs(vals)
    if raw:
        tensor.raw_data = vals
    else:
        field = mapping.STORAGE_TENSOR_TYPE_TO_FIELD[
            mapping.TENSOR_TYPE_TO_STORAGE_TENSOR_TYPE[data_type]]
        getattr(tensor, field).extend(vals)

    tensor.dims.extend(dims)
    return tensor


def get_opset_number_from_onnx():
    """
    Returns the latest opset version supported
    by the *onnx* package.
    """
    return defs.onnx_opset_version()


def get_latest_tested_opset_version():
    """
    This module relies on *onnxruntime* to test every
    converter. The function returns the most recent
    target opset tested with *onnxruntime* or the opset
    version specified by *onnx* package if this one is lower
    (return by `onnx.defs.onnx_opset_version()`).
    """
    from .. import __max_supported_opset__
    return min(__max_supported_opset__, get_opset_number_from_onnx())



# SPDX-License-Identifier: Apache-2.0

from logging import getLogger
from collections import OrderedDict
import numpy as np
from ..common.data_types import FloatTensorType, DoubleTensorType
from ..common.utils import get_unique_subgraph
from .onnx_ops import (
    OnnxIdentity, OnnxScan, OnnxTranspose,
    OnnxSub, OnnxReduceSumSquare,
    OnnxSqrt, OnnxPow, OnnxAbs, OnnxReduceSumApi11)

logger = getLogger('skl2onnx')


def onnx_squareform_pdist(X, metric='sqeuclidean', dtype=None,
                          op_version=None, **kwargs):
    """
    Returns the ONNX graph which computes
    ``squareform(pdist(X, metric=metric))``.
    """
    if metric == 'sqeuclidean':
        return _onnx_squareform_pdist_sqeuclidean(
            X, dtype=dtype, op_version=op_version, **kwargs)
    if metric == 'euclidean':
        res = _onnx_squareform_pdist_sqeuclidean(
            X, dtype=dtype, op_version=op_version)
        return OnnxSqrt(res, op_version=op_version, **kwargs)
    raise NotImplementedError(
        "metric='{}' is not implemented.".format(metric))


def _onnx_squareform_pdist_sqeuclidean(X, dtype=None, op_version=None,
                                       **kwargs):
    """
    Returns the ONNX graph which computes
    ``squareform(pdist(X, metric='sqeuclidean'))``.
    """
    unique = get_unique_subgraph()
    diff = OnnxSub('next_in', 'next',
                   op_version=op_version)
    id_next = OnnxIdentity('next_in', output_names=['next_out'],
                           op_version=op_version)
    flat = OnnxReduceSumSquare(diff, axes=[1], op_version=op_version,
                               output_names=['scan_out'], keepdims=0)
    flat.set_onnx_name_prefix('cflat_%d' % unique)
    id_next.set_onnx_name_prefix('pdistsqe_%d' % unique)
    tensor_type = FloatTensorType if dtype == np.float32 else DoubleTensorType
    scan_body = id_next.to_onnx(
        OrderedDict([('next_in', tensor_type([None, None])),
                     ('next', tensor_type([None]))]),
        outputs=[('next_out', tensor_type([None, None])),
                 ('scan_out', tensor_type([None]))],
        other_outputs=[flat],
        target_opset=op_version)

    node = OnnxScan(X, X, output_names=['u(scan0)', 'u(scan1)'],
                    num_scan_inputs=1,
                    body=(scan_body.graph, [id_next, flat]),
                    op_version=op_version, **kwargs)
    logger.debug('[_onnx_squareform_pdist_sqeuclidean] +Scan dtype=%r',
                 dtype)
    return node[1]


def onnx_cdist(XA, XB, metric='sqeuclidean', dtype=None,
               op_version=None, dim_in=None, dim_out=None,
               **kwargs):
    """
    Returns the ONNX graph which computes
    ``cdist(XA, XB, metric=metric)``.

    :param XA: array or OnnxOperatorMixin
    :param XB: array or OnnxOperatorMixin
    :param metric: distance type
    :param dtype: *np.float32* or *np.float64*
    :param op_version: opset version
    :param dim_in: dimension of the input vectorial space
        (if known)
    :param dim_out: dimension of the output vectorial space
        (if known)
    :param kwargs: addition parameter
    :return: OnnxOperatorMixin
    """
    if metric == 'sqeuclidean':
        return _onnx_cdist_sqeuclidean(
            XA, XB, dtype=dtype, op_version=op_version,
            dim_in=dim_in, dim_out=dim_out, **kwargs)
    elif metric == 'euclidean':
        res = _onnx_cdist_sqeuclidean(
            XA, XB, dtype=dtype, op_version=op_version,
            dim_in=dim_in, dim_out=dim_out)
        return OnnxSqrt(res, op_version=op_version, **kwargs)
    elif metric == 'minkowski':
        p = kwargs.pop('p')
        res = _onnx_cdist_minkowski(
            XA, XB, dtype=dtype, op_version=op_version, p=p,
            dim_in=dim_in, dim_out=dim_out)
        return OnnxPow(res, np.array([1. / p], dtype=dtype),
                       op_version=op_version, **kwargs)
    elif metric in ('manhattan', 'cityblock'):
        return _onnx_cdist_manhattan(
            XA, XB, dtype=dtype, op_version=op_version,
            dim_in=dim_in, dim_out=dim_out, **kwargs)
    else:
        raise NotImplementedError("metric='{}' is not implemented.".format(
            metric))


def _onnx_cdist_begin(op_version):
    diff = OnnxSub('next_in', 'next',
                   op_version=op_version)
    id_next = OnnxIdentity('next_in', output_names=['next_out'],
                           op_version=op_version)
    return diff, id_next


def _onnx_cdist_end(XA, XB, id_next, flat, dtype, op_version,
                    dim_in=None, dim_out=None, **kwargs):
    unique = get_unique_subgraph()
    tensor_type = FloatTensorType if dtype == np.float32 else DoubleTensorType
    id_next.set_onnx_name_prefix('cdistd_%d' % unique)
    flat.set_onnx_name_prefix('cdistdf_%d' % unique)
    shape_in = (tensor_type([None, None]) if dim_in is None
                else tensor_type([None, dim_in]))
    scan_body = id_next.to_onnx(
        OrderedDict([('next_in', shape_in),
                     ('next', tensor_type([None]))]),
        outputs=[('next_out', tensor_type([None, None])),
                 ('scan_out', tensor_type([None]))],
        other_outputs=[flat],
        target_opset=op_version)
    logger.debug('[_onnx_cdist_end] + Scan dim_in=%r dim_out=%r dtype=%r',
                 dim_in, dim_out, dtype)

    node = OnnxScan(XA, XB, output_names=['u(scan0)', 'u(scan1)'],
                    num_scan_inputs=1,
                    body=(scan_body.graph, [id_next, flat]),
                    op_version=op_version)
    return OnnxTranspose(node[1], perm=[1, 0], op_version=op_version,
                         **kwargs)


def _onnx_cdist_sqeuclidean(XA, XB, dtype=None, op_version=None,
                            dim_in=None, dim_out=None, **kwargs):
    """
    Returns the ONNX graph which computes
    ``cdist(X, metric='sqeuclidean')``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    norm = OnnxReduceSumSquare(
        diff, axes=[1], keepdims=0, op_version=op_version)
    flat = OnnxIdentity(norm, output_names=['scan_out'], op_version=op_version)
    return _onnx_cdist_end(XA, XB, id_next, flat, dtype, op_version,
                           dim_in=dim_in, dim_out=dim_out, **kwargs)


def _onnx_cdist_minkowski(XA, XB, dtype=None, op_version=None, p=2,
                          dim_in=None, dim_out=None, **kwargs):
    """
    Returns the ONNX graph which computes the Minkowski distance
    or ``minkowski(XA, XB, p)``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    diff_pow = OnnxPow(OnnxAbs(diff, op_version=op_version),
                       np.array([p], dtype=dtype), op_version=op_version)
    norm = OnnxReduceSumApi11(
        diff_pow, axes=[1], keepdims=0, op_version=op_version)
    norm.set_onnx_name_prefix("norm_%d" % id(norm))
    flat = OnnxIdentity(norm, output_names=['scan_out'], op_version=op_version)
    return _onnx_cdist_end(XA, XB, id_next, flat, dtype, op_version,
                           dim_in=dim_in, dim_out=dim_out, **kwargs)


def _onnx_cdist_manhattan(XA, XB, dtype=None, op_version=None,
                          dim_in=None, dim_out=None, **kwargs):
    """
    Returns the ONNX graph which computes the Manhattan distance
    or ``Manhattan(X, Y)``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    diff_pow = OnnxAbs(diff, op_version=op_version)
    norm = OnnxReduceSumApi11(
        diff_pow, axes=[1], keepdims=0, op_version=op_version)
    norm.set_onnx_name_prefix("norm_%d" % id(norm))
    flat = OnnxIdentity(norm, output_names=['scan_out'], op_version=op_version)
    return _onnx_cdist_end(XA, XB, id_next, flat, dtype, op_version,
                           dim_in=dim_in, dim_out=dim_out, **kwargs)
